from pydantic import validate_arguments

from kmm import CarDirection
from kmm.positions.positions import Positions


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def wire_camera_positions(positions: Positions, direction: CarDirection):
    ascending = kmm_ascending(positions.dataframe)

    if (
        (direction == CarDirection.A and ascending)
        or (direction == CarDirection.B and not ascending)
    ):
        correction = -8

    elif (
        (direction == CarDirection.A and not ascending)
        or (direction == CarDirection.B and ascending)
    ):
        correction = 8

    else:
        raise ValueError

    return positions.replace(
        dataframe=(
            positions.dataframe
            .assign(meter=lambda df: df["meter"] + correction)
        )
    )


def kmm_ascending(dataframe):
    total_meter = dataframe["kilometer"] * 1000 + dataframe["meter"]
    diff = total_meter.values[:-1] - total_meter.values[1:]
    descending = (diff < 0).mean()
    ascending = (diff > 0).mean()

    if descending < 0.9 and ascending < 0.9:
        raise ValueError("Unable to determine ascending/descending kmm numbers")

    else:
        return ascending > 0.9


def test_camera_positions_kmm():
    from kmm import Header

    positions = Positions.from_path("tests/ascending_B.kmm")
    header = Header.from_path("tests/ascending_B.hdr")
    positions_calibrated = wire_camera_positions(positions, header.car_direction)
    assert (
        positions_calibrated.dataframe["meter"].iloc[0]
        == positions.dataframe["meter"].iloc[0] - 8
    )


def test_camera_positions_kmm2():
    from kmm import Header

    positions = Positions.from_path("tests/ascending_B.kmm2")
    header = Header.from_path("tests/ascending_B.hdr")
    positions_calibrated = wire_camera_positions(positions, header.car_direction)
    assert (
        positions_calibrated.dataframe["meter"].iloc[0]
        == positions.dataframe["meter"].iloc[0] - 8
    )



from pathlib import Path
from xml.etree import ElementTree
from pydantic import validate_arguments

import kmm


class Header(kmm.FunctionalBase):
    car_direction: kmm.CarDirection
    position: int
    sync: int

    @staticmethod
    @validate_arguments
    def from_path(path: Path):
        """
        Loads header data from .hdr file.
        """
        tree = ElementTree.parse(path)
        position, sync = kmm.header.position_sync(tree)
        return Header(
            position=position,
            sync=sync,
            car_direction=kmm.header.car_direction(tree),
        )


def test_header():
    Header.from_path("tests/ascending_B.hdr")


def test_empty_header():
    Header.from_path("tests/empty.hdr")



import re
from xml.etree import ElementTree


def position_sync(tree: ElementTree):

    root = tree.getroot()
    sync_tags = [
        child.text
        for child in root
        if child.tag == "Sync"
    ]

    if len(sync_tags) == 0:
        raise ValueError("Did not find a sync tag in header.")

    sync_tag = sync_tags[0]

    position = int(re.search(
        r"Position = \"(\d*)\"",
        sync_tag,
    ).group(1))

    sync = int(re.search(
        r"Sync = \"(\d*)\"",
        sync_tag,
    ).group(1))

    return position, sync



from pydantic import BaseModel, Extra


class FunctionalBase(BaseModel):
    class Config:
        allow_mutation = False
        extra = Extra.forbid

    def map(self, fn, *args, **kwargs):
        return fn(self, *args, **kwargs)

    def replace(self, **kwargs):
        new_dict = self.dict()
        new_dict.update(**kwargs)
        return type(self)(**new_dict)



from pathlib import Path
import pandas as pd
from pydantic import validate_arguments

import kmm


class Positions(kmm.FunctionalBase):

    dataframe: pd.DataFrame

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    @validate_arguments
    def from_path(path: Path):
        """
        Loads positions from .kmm or .kmm2 file.
        """
        if path.suffix == ".kmm":
            dataframe = kmm.positions.read_kmm(path)
        elif path.suffix == ".kmm2":
            dataframe = kmm.positions.read_kmm2(path)
        else:
            raise ValueError(f"Unable to parse file type {path.suffix}")

        return Positions(dataframe=dataframe)

    @staticmethod
    @validate_arguments
    def read_sync_adjust(
        kmm_path: Path,
        header_path: Path,
        adjustment: kmm.PositionAdjustment = kmm.PositionAdjustment.WIRE_CAMERA,
    ):
        """
        Loads positions from .kmm or .kmm2 file + .hdr file, then performs
        frame index sync, position adjustment and geodetic coordinate transformation.
        """
        header = kmm.Header.from_path(header_path)
        return (
            Positions.from_path(kmm_path)
            .sync_frame_index(header)
            .adjust(adjustment, header)
            .geodetic()
        )

    def sync_frame_index(self, header):
        return self.replace(
            dataframe=self.dataframe.assign(
                frame_index=(
                    (
                        self.dataframe["centimeter"]
                        + header.position
                        - header.sync
                    ) / 10
                ).astype(int)
            )
        )

    def adjust(self, adjustment: kmm.PositionAdjustment, header):
        if adjustment == kmm.PositionAdjustment.WIRE_CAMERA:
            return kmm.positions.wire_camera_positions(self, header.car_direction)
        else:
            raise ValueError(f"Unknown adjustment option {adjustment}")

    def geodetic(self):
        return kmm.positions.geodetic(self)


def test_read_kmm():
    positions = Positions.read_sync_adjust(
        "tests/ascending_B.kmm", "tests/ascending_B.hdr"
    )
    assert len(positions.dataframe) > 0


def test_read_kmm2():
    positions = Positions.read_sync_adjust(
        "tests/ascending_B.kmm2", "tests/ascending_B.hdr"
    )
    assert len(positions.dataframe) > 0


def test_empty_kmm():
    positions = Positions.from_path("tests/empty.kmm")
    assert len(positions.dataframe) == 0


def test_empty_kmm2():
    positions = Positions.from_path("tests/empty.kmm2")
    assert len(positions.dataframe) == 0



from kmm.functional_base import FunctionalBase
from kmm.car_direction import CarDirection
from kmm.position_adjustment import PositionAdjustment
from kmm import positions
from kmm import header
from kmm.positions.positions import Positions
from kmm.header.header import Header



import re
from xml.etree import ElementTree

import kmm


def car_direction(tree: ElementTree):

    root = tree.getroot()
    start_tags = [
        child.text
        for child in root
        if child.tag == "Start"
    ]

    if len(start_tags) != 1:
        raise ValueError(f"Expected 1 'Start' tag in header, found {len(start_tags)}")

    start_tag = start_tags[0]
    car_direction = re.search(
        r"CarDirection = \"(.*)\"",
        start_tag,
    ).group(1)

    if not any(car_direction == item.value for item in kmm.CarDirection):
        raise ValueError(f"Unknown car direction {car_direction}")

    return kmm.CarDirection[car_direction]



from sweref99 import projections

from kmm.positions.positions import Positions


tm = projections.make_transverse_mercator("SWEREF_99_TM")


def geodetic(positions: Positions):
    dataframe = positions.dataframe
    if len(dataframe) == 0:
        dataframe = dataframe.assign(longitude=[], latitude=[])
    else:
        latitude, longitude = zip(*dataframe.apply(
            lambda row: (tm.grid_to_geodetic(row["sweref99_tm_x"], row["sweref99_tm_y"])),
            axis="columns",
            result_type="reduce",
        ))
        dataframe = dataframe.assign(longitude=longitude, latitude=latitude)
    return positions.replace(dataframe=dataframe)


def test_geodetic():

    positions = Positions.from_path("tests/ascending_B.kmm2")
    df = geodetic(positions).dataframe
    assert ((df["latitude"] < 68) & (df["latitude"] > 55)).all()
    assert ((df["longitude"] < 25) & (df["longitude"] > 7)).all()



from kmm.positions.read_kmm import read_kmm
from kmm.positions.read_kmm2 import read_kmm2
from kmm.positions.geodetic import geodetic
from kmm.positions.wire_camera_positions import wire_camera_positions
from kmm.positions.positions import Positions
